"""

The packetlogic2 package contains the multiversion PacketLogic API.

This module provides a single entry point into the PacketLogic
PythonAPI, the (L{packetlogic2.connect() <connect>}) function.

   >>> import packetlogic2
   >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
   >>> print pl.remoteversion
   12.1.2.70
   >>> rt = pl.Realtime()
   >>> rs = pl.Ruleset()

"""

class _UnimplementedException:
	""" Used as placeholder for exceptions not implemented in certain versions,
	to allow writing try-except-clauses in multi-version environments, without a lot
	of if:s and stuff, even if that particular except-handler will never run.
	"""
	pass

def _available_apis():
	import availableapis
	
	return ["v%s" % x for x in availableapis.available_apis.split()]

def _import_api(api):
	return __import__('packetlogic2.%s' % api, globals(), locals(), ['*'])

def _find_version(host, username, password, port, versions, timeout):
	if not versions:
		versions = _available_apis()
		
	bad_version_msg = ''
	for api in versions:
		tmp_api = _import_api(api)
		try:
			tmp_connection = PLConnection(host, username, password, port, api, tmp_api, timeout)
			r = tmp_connection._check_connection()
			tmp_version = None
			try:
				tmp_version = r.version[1]
			except:
				pass
			r.close()
			
		except tmp_api.pldb.PLDBError, msg:
			# Quench all Bad versions, or convert any other version-specific PLDBError to RuntimeError
			if not str(msg).startswith("Bad version"):
				raise RuntimeError("PLDBError: %s" % str(msg))

			else:
				bad_version_msg = ': ' + str(msg)
			
		except RuntimeError, msg:
			# Quench all Bad versions
			if not str(msg).startswith("Bad version"):
				raise
			else:
				bad_version_msg = ': ' + str(msg)
		else:
			return api, tmp_version

	raise RuntimeError("Could not determine server version" + bad_version_msg)

def _deprecated_exc(val):
	def _(self):
		import warnings
		warnings.warn("Usage of PLConnection.%s is deprecated, use the packetlogic2.exceptions module" % val, DeprecationWarning, stacklevel=2)

		import packetlogic2.exceptions
		return getattr(packetlogic2.exceptions, val)
	doc = """
	This is a deprecated way to access the %(name)s exception.

	@deprecated: Use the packetlogic2.exceptions module instead.
	""" % {'name': val}
	return property(_, doc=doc)
	

class PLConnection(object):
	"""
	Returned by the L{connect() <connect>} function. Wrapper for communication with a PacketLogic system.

	This object does not keep any (TCP-level) connection to the
	PacketLogic system, it only stores the connection
	information. By invoking the different resource methods
	(e.g. Ruleset(), Realtime()) a new connection is made to the
	PacketLogic.

	@ivar username: The username used to connect to the PacketLogic system (The <username> argument).
	"""
	
	def __init__(self, host, username, password, port, plversion, api, timeout, remoteversion = "Unknown"):
		"""
		This class should not be instantiated directly, use the L{connect() <connect>} function.
		"""
		# Save public attributes
		self._host = host
		self.username = username
		self._plversion = plversion
		self._remoteversion = remoteversion

		# Save internal attributes
		self.password = password
		self.api = api
		self._pld_port = 42001
		if port is None:
			self._pldb_port = 42000
		else:
			self._pldb_port = port

		self.kwargs = {}
		if hasattr(self.api.pldb.Ruleset, 'SUPPORTS_TIMEOUT'):
			self.kwargs['timeout'] = timeout

	remoteversion = property(lambda self: self._remoteversion, doc="The version of the firmware running on the the PacketLogic system.")
	plversion = property(lambda self: self._plversion, doc="Protocolversion of connected PacketLogic system, and hence the API used in this connection.")
	host = property(lambda self: self._host, doc="The hostname or IP address of the PacketLogic system (The <host> argument).")
	
	PL_API_VERSION = property(lambda self: self.api.pldb.PL_API_VERSION, doc="The full API version string.")

	PLDBError = _deprecated_exc("PLDBError")
	PLDError = _deprecated_exc("PLDError")
	PLDPermissionError = _deprecated_exc("PLDPermissionError")
	PLDNotAvailable = _deprecated_exc("PLDNotAvailable")
	PLUnsupportedFeatureError = _deprecated_exc("PLUnsupportedFeatureError")
	PLDServerError = _deprecated_exc("PLDServerError")

	StatsZone = property(lambda self: self.api.pld.StatsZone, doc="StatsZone class extracted from the specific API used in this connection.")

	def __repr__(self):
		return '<PLConnection to %s@%s version %s>' % (self.username, self._host, self._plversion)

	def _check_connection(self):
		return self.api.pldb.ConnectChecker(self._host, self.username, self.password, self._pldb_port, **self.kwargs)

	def Ruleset(self):
		"""
		Get a connection to the PacketLogic's Ruleset resource.

		For details about the Ruleset resource see
		documentation for the L{Ruleset object
		<ruleset.Ruleset>}.

		For details about the resource see documentation for the L{Ruleset object <ruleset.Ruleset>}.

		@rtype: L{Ruleset object <ruleset.Ruleset>}
		"""
		return self.api.pldb.Ruleset(self._host, self.username, self.password, self._pldb_port, **self.kwargs)
	
	def Statistics(self):
		"""
		Get a connection to the PacketLogic's Statistics resource.

		For details about the Statistics resource see
		documentation for the L{Statistics object
		<statistics.Statistics>}.

		@rtype: L{Statistics object <statistics.Statistics>}
		"""
		return self.api.pldb.Statistics(self._host, self.username, self.password, self._pldb_port, **self.kwargs)
	
	def Config(self):
		"""
		Get a connection to the PacketLogic's Config resource.

		For details about the Config resource see
		documentation for the L{Config object <config.Config>}.

		@rtype: L{Config object <config.Config>}
		"""
		return self.api.pldb.Config(self._host, self.username, self.password, self._pldb_port, **self.kwargs)
	
	def ConnLog(self):
		"""
		Get a connection to the PacketLogic's ConnLog resource.

		For details about the ConnLog resource see
		documentation for the L{ConnLog object
		<connlog.ConnLog>}.

		@rtype: L{ConnLog object <connlog.ConnLog>}
		"""
		return self.api.pldb.ConnLog(self._host, self.username, self.password, self._pldb_port, **self.kwargs)
	
	def ResourceEdit(self, port=None):
		"""
		Get a connection to the PacketLogic's ResourceEdit resource.

		For details about the ResourceEdit resource see
		documentation for the L{ResourceEdit object
		<resourceedit.ResourceEdit>}.

		@rtype: L{ResourceEdit object <resourceedit.ResourceEdit>}
		"""
		return self.api.pldb.ResourceEdit(self._host, self.username, self.password, self._pldb_port, **self.kwargs)

	def Backup(self, port=None):
		"""
		Get a connection to the PacketLogic's Backup resource.

		For details about the Backup resource see
		documentation for the L{Backup object <backup.Backup>}.

		@rtype: L{Backup object <backup.Backup>}
		"""
		return self.api.pldb.Backup(self._host, self.username, self.password, self._pldb_port, **self.kwargs)

	def SystemOverview(self):
		"""
		Get a connection to the PacketLogic's SystemOverview resource.

		For details about the SystemOverview resource see
		documentation for the L{SystemOverview object
		<systemoverview.SystemOverview>}.

		@rtype: L{SystemOverview object <systemoverview.SystemOverview>}
		"""
		return self.api.pldb.SystemOverview(self._host, self.username, self.password, self._pldb_port, **self.kwargs)

	def Fileserv(self):
		"""
		Get a connection to the PacketLogic's Fileserv resource.

		For details about the Fileserv resource see
		documentation for the L{Fileserv object
		<fileserv.Fileserv>}.

		@rtype: L{Fileserv object <fileserv.Fileserv>}
		"""
		return self.api.pldb.Fileserv(self._host, self.username, self.password, self._pldb_port, **self.kwargs)

	def Channels(self):
		"""
		Get a connection to the PacketLogic's channels resource.

		For details about the Channels resource see
		documentation for the L{Channels object
		<channels.Channels>}.

		@rtype: L{Channels object <channels.Channels>}
		"""
		return self.api.pldb.Channels(self._host, self.username, self.password, self._pldb_port, **self.kwargs)

	def Users(self):
		"""
		Get a connection to the PacketLogic's users resource.

		For details about the Users resource see
		documentation for the L{Users object
		<user.Users>}.

		@rtype: L{Users object <user.Users>}
		"""
		return self.api.pldb.Users(self._host, self.username, self.password, self._pldb_port, **self.kwargs)

	def Realtime(self, port=None):
		"""
		Get a connection to the PacketLogic's realtime resource.

		For details about the Realtime resource see
		documentation for the L{Realtime object <pld.PLD>}.

		@rtype: L{Realtime object <pld.PLD>}
		"""
		return self.api.pld.PLD(self._host, self.username, self.password, port or self._pld_port, **self.kwargs)

	def HostTriggers(self):
		"""
		Get a connection to the PacketLogic's host trigger resource.

		For details about the Host trigger resource see documentation
		for the L{HostTriggers object <pldb.HostTriggers>}.

		@rtype: L{HostTriggers object <pldb.HostTriggers>}
		"""
		return self.api.pldb.HostTriggers(self._host, self.username, self.password, self._pldb_port, **self.kwargs)

	def SysDiag(self):
		"""
		Get a connection to the PacketLogic's System Diagnostics resource.

		For details about the System Diagnostics resource see documentation
		for the L{SysDiag object <pldb.SysDiag>}.

		@rtype: L{SysDiag object <pldb.SysDiag>}

		@attention: This only exist in firmware version 13.1 and newer.
		"""
		return self.api.pldb.SysDiag(self._host, self.username, self.password, self._pldb_port, **self.kwargs)

	def ConnprotTriggers(self):
		"""
		Get a connection to the PacketLogic's connection protection trigger resource.

		For details about the connection protection trigger resource see documentation
		for the L{ConnprotTriggers object <pldb.ConnprotTriggers>}.

		@rtype: L{ConnprotTriggers object <pldb.ConnprotTriggers>}
		"""
		return self.api.pldb.ConnprotTriggers(self._host, self.username, self.password, self._pldb_port, **self.kwargs)

	def _DynamicLiveView(self):
		""" Returns a L{DynamicLiveView object <pldb.DynamicLiveView>}."""
		return self.api.pldb.DynamicLiveView(self._host, self.username, self.password, self._pldb_port, **self.kwargs)

	def _SessionContext(self):
		"""
		Get a connection to the PacketLogic's Session Context resource.

		For details about the session context resource see documentation
		for the L{SessionContext object <plscd.PLSCD>}.

		@rtype: L{SessionContext object <plscd.PLSCD>}
		"""
		return self.api.plscd.PLSCD(self._host, self.username, self.password, **self.kwargs)

def connect(host, username, password, port=None, versions=None, timeout=None):
	"""
	Returns a L{PLConnection object <PLConnection>} using the appropriate protocol version.

	@type host: C{str}
	@param host: PacketLogic host or IP address to connect to

	@type username: C{str}
	@param username: User to authenticate as

	@type password: C{str}
	@param password: Password to authenticate with

	@type port: C{int}
	@param port: Connect to specified port instead of default port

	@type versions: C{list}
	@param versions: Used to specify explicit order of versions to try.
	

	@type timeout: C{float}
	@param timeout: Timeout in seconds to use for connecting and performing actions.

	@raises IOError: If connection to system fails.
	@raises RuntimeError: B{Couldn't log in}: When authentication fails.
	"""
	if versions is not None and (not isinstance(versions, list) or not versions):
		raise TypeError("argument is netither None, nor a list of version(s)")
	
	v, ver = _find_version(host, username, password, port, versions, timeout)
	return PLConnection(host, username, password, port, v, _import_api(v), timeout, ver)
