"""
This module contains IpfixObject implementation.

There is no need to import this module directly. Use the methods of
the L{Ruleset object <ruleset.Ruleset>} to create and manipulate
IpfixObjects.

    >>> o = r.ipfixobject_add('Test Ipfix')

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

import plutil

from packetlogic2.exceptions import PLDBUnsupportedInFirmware

def _ipfixobjectlist(pldb):
	cid = pldb.session.ipfix_object_list()
	return plutil._do(pldb, cid, plutil._ListCB(pldb, pldb.session.ipfix_object_list_handle, IpfixObject))

def _ipfixobjectadd(_pldb, name, templatestring):
	kwargs = locals().copy()

	cid = _pldb.session.ipfix_object_add(name, templatestring)
	(id, tstamp) = plutil._do(_pldb, cid, plutil._AddCB(_pldb, _pldb.session.ipfix_object_add_handle))
	res = IpfixObject(id = id,
			       creation_date = tstamp,
			       modification_date = tstamp,
			       creator = _pldb.user,
			       modified_by = _pldb.user,
			       **kwargs)
	return res

def _rename(pldb, obj, newname):
	cid = pldb.session.ipfix_object_rename(obj.id, newname)
	return plutil._do(pldb, cid, plutil._AddCB(pldb, pldb.session.ipfix_object_rename_handle))

def _remove(pldb, obj):
	cid = pldb.session.ipfix_object_remove(obj.id)
	return plutil._do(pldb, cid, plutil._AddCB(pldb, pldb.session.ipfix_object_remove_handle))

_PropatisticsObject = plutil.Prop('id', 'name', 'templatestring',
				  'creator', 'creation_date', 'modified_by',
				  'modification_date')


class IpfixObject(object):
	"""
	Base class for IpfixObjects.

	This class should not be instantiated and should instead be created and
	accessed through:

		- L{Ruleset.ipfixobject_add}
		- L{Ruleset.ipfixobject_find}
		- L{Ruleset.ipfixobject_find_id}
		- L{Ruleset.ipfixobject_list}
		- L{Ruleset.ipfixobject_remove}

	Class variables are read-only, use methods to alter the object.

	"""
	__slots__ = ('_prop', '_pldb')

	def __init__(self, _pldb=None, **props):
		self._pldb = _pldb
		self._prop = _PropatisticsObject(**props)

	def __repr__(self):
		return "<PacketLogic IpfixObject '%s'>" % self._prop.name

	def rename(self, newname):
		"""
		Rename the current object

		I{destroys netobject cache}

		@type newname: C{str}
		@param newname: The name that the object should have after rename process

		"""
		_rename(self._pldb, self, newname)
		self._prop.name = newname

		self._pldb._invalidate_cache()

	def _machineid(self):
		""" Use of property 'machineid' is deprecated since v11.0beta5, use 'systemid' instead."""
		import warnings
		warnings.warn("Use of property 'machineid' is deprecated, use 'systemid' instead.",
			      DeprecationWarning, stacklevel=2)
		return self.systemid

	def change(self, *args, **kvargs):
		raise NotImplementedError("Modification of ipfixobjects not implemented yet")

	machineid = property(_machineid)

	id = property(lambda self: self._prop.id, doc="id number of this object")
	name = property(lambda self: self._prop.name, doc="Name of the ipfix object")

	templatestring = property(lambda self: self._prop.templatestring, doc="Template string")

	creation_date = property(lambda self: self._prop.creation_date, doc="Date and time when this object was created")
	modification_date = property(lambda self: self._prop.modification_date, doc="Date and time when this object was last modified")
	creator = property(lambda self: self._prop.creator, doc="Username that created this object")
	modified_by = property(lambda self: self._prop.modified_by, doc="Username that last modified this object")

	systemid = plutil._sysidprop("System ID of the PacketLogic this object belongs to. (Replaces machineid)")
	is_global = property(lambda self: False, doc="True if this object resides on another system (always false in 12.1)")
