"""
This module contains functionality for managing system configuration
on PacketLogic system.

Do NOT import this module directly. The L{PLConnection
<packetlogic2.PLConnection>} object should be used, e.g:

    >>> import packetlogic2
    >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
    >>> c = pl.Config()

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

from resource import Resource
from thisversion import PLDB

import plutil
import warnings

class Config(Resource):
	""" This is the resource for manipulating configuration.

	All configuration options are returned as a string representing
	their value. When changing a value you should use the actual
	corresponding python type though. The Config.TYPE_ constants
	are mapped to the number returned in the 'type' item by the
	L{list} method. Their name and the below help tells you which
	python type and which format that configuration option uses.

	@cvar TYPE_INVISIBLE: Can't be set.
	@cvar TYPE_STRING: C{str}
	@cvar TYPE_INTEGER: C{int}
	@cvar TYPE_IPADDRESS: IP address as C{str}
	@cvar TYPE_FLOAT: C{float}
	@cvar TYPE_BOOLEAN: C{bool}

	@group Config: list,set,get
	@group Misc: reload,reboot
	"""
	
	TYPE_INVISIBLE = PLDB.CONFIG_TYPE_INVISIBLE
	TYPE_STRING = PLDB.CONFIG_TYPE_STRING
	TYPE_INTEGER = PLDB.CONFIG_TYPE_INTEGER
	TYPE_IPADDRESS = PLDB.CONFIG_TYPE_IPADDRESS
	TYPE_FLOAT = PLDB.CONFIG_TYPE_FLOAT
	TYPE_BOOLEAN = PLDB.CONFIG_TYPE_BOOLEAN

	_items = None
        _flags = None
        _categories = None

	def _resource(self):
		return PLDB.RESOURCE_CONFIG

	_VERIFY_MINMAX = True

	def reload(self, message):
		"""
		Reload configuration.

		This will reload the PacketLogic configuration, and will cause the API to get disconnected.

		@param message: Message describing the reason for the reload.
		@type message: C{str}
		
		"""
		return plutil._do(self, self.session.reload(message), plutil._AddCB(self, self.session.reloading_handle))

	def reboot(self, message):
		"""
		Reboot PacketLogic.

		This will reboot the PacketLogic hardware.
		
		@param message: Message describing the reason for the reload.
		@type message: C{str}
		
		"""
		return plutil._do(self, self.session.restart(message), plutil._AddCB(self, self.session.restarting_handle))

	def set(self, key, value):
		"""
		Set configuration value

			>>> x.set("PACKET_POOL_SIZE", 500000)
			>>> x.set("DRDL_ENABLED", True)
			>>> x.set("BGP_SERVER", "192.168.0.47")

		@param key: Name of configuration option. The available
		            options can be seen by the L{list} method.
		@type key: C{str}
		
		@param value: New value for the configuration option.
		              The type depends on the variable. Use the
			      'type' item in the dict returned by the
			      L{list} method in conjunction with the
			      Config.TYPE_* constants to decide type and
			      format of each configuration option.
			      
		@type value: C{str}, C{int}, C{float} or C{bool}.

		"""

		if not self._items or key not in self._items:
			self.list()

		if key not in self._items:
			raise ValueError("Config key %s not found" % key)

		if self._items[key]['type'] == PLDB.CONFIG_TYPE_INVISIBLE:
			raise ValueError("Can't modify invisible values")
					
		elif self._items[key]['type'] == PLDB.CONFIG_TYPE_STRING:
			if not isinstance(value, str):
				raise TypeError("Config key %s requires string argument" % (key))
						
		elif self._items[key]['type'] == PLDB.CONFIG_TYPE_INTEGER:
			if not isinstance(value, (int, long)):
				raise TypeError("Config key %s requires int argument" % (key))
			elif value > int(self._items[key]['maxvalue']):
				if self._VERIFY_MINMAX:
					raise ValueError("Value is bigger than maximum allowed value %d" % int(self._items[key]['maxvalue']))
				warnings.warn("Setting out-of-range configuration value. This is not supported and may render system unusable")
			elif value < int(self._items[key]['minvalue']):
				if self._VERIFY_MINMAX:
					raise ValueError("Value is smaller than minimum allowed value %d" % int(self._items[key]['minvalue']))
				warnings.warn("Setting out-of-range configuration value. This is not supported and may render system unusable")

			value = str(value)
				
		elif self._items[key]['type'] == PLDB.CONFIG_TYPE_IPADDRESS:
			if not isinstance(value, str):
				raise TypeError("Config key %s requires string argument" % (key))
						
			if not plutil.verify_ip46(value):
				raise ValueError("Config key %s requires an IP address" % (key))
					
		elif self._items[key]['type'] == PLDB.CONFIG_TYPE_FLOAT:
			if not isinstance(value, float):
				raise TypeError("Config key %s requires float argument" % (key))
			elif value > int(self._items[key]['maxvalue']):
				if self._VERIFY_MINMAX:
					raise ValueError("Value is bigger than maximum allowed value %f" % float(self._items[key]['maxvalue']))
				warnings.warn("Setting out-of-range configuration value. This is not supported and may render system unusable")
			elif value < int(self._items[key]['minvalue']):
				if self._VERIFY_MINMAX:
					raise ValueError("Value is smaller than minimum allowed value %f" % float(self._items[key]['minvalue']))
				warnings.warn("Setting out-of-range configuration value. This is not supported and may render system unusable")

			value = str(value)
				
		elif self._items[key]['type'] == PLDB.CONFIG_TYPE_BOOLEAN:
			if not isinstance(value, bool):
				raise TypeError("Config key %s requires boolean argument" % (key))
			else:
				# Cast boolean True/False to 1/0
				value = str(value+0)
		else:
			raise ValueError("Config key %s has unknown type %d" % (key, self._items[key]['type']))

		self.session.config_set(key, value)

	def get(self, key):
		"""
		Get information about a specific config option.

		This method is equivalent to list()[key]

		@param key: Name of configuration option. The available
		            options can be seen by the L{list} method.
		@type key: C{str}

		@rtype: C{dict}
		@returns: A dict describing a config option
		"""
		if not self._items or key not in self._items:
			self.list()

		if key not in self._items:
			raise ValueError("Config key %s not found" % key)

		return self._items[key]

	def getvalue(self, key):
		"""
		Get the value of a specific config option.

		This method is equivalent to list()[key]['value']

		@param key: Name of configuration option.
		@type key: C{str}

		@returns: The value of the requested config option.
		"""
		if not self._items or key not in self._items:
			self.list()

		if key not in self._items:
			raise ValueError("Config key %s not found" % key)

		return self._items[key]['value']

        def categories(self):
                """
                Retrieve a mapping of bitnumber to category name.
                """
                if self._categories is None:
                        self._categories = plutil._do(self, self.session.config_category_list(), plutil._ListCB(self, self.session.config_category_list_handle, None, False))
                return self._categories

        def flags(self):
                """
                Retrieve a mapping of bitnumber to flag name.
                """
                if self._flags is None:
                        self._flags = plutil._do(self, self.session.config_flags_list(), plutil._ListCB(self, self.session.config_flags_list_handle, None, False))
                return self._flags

	def list(self):
		"""
		List configvalues.
		The values are always represented by a string, but
		the actual value is different for different configuration
		options. The 'type' item in conjunction with the
		Config.TYPE_* constants can be used to decide the type.

			>>> x.list()[:2]
			[{'defvalue': '',
			  'description': 'Lowlevel accept filter',
			  'key': 'ACCEPT_FILTERS',
			  'maxvalue': '',
			  'minvalue': '',
			  'type': 1,
			  'value': '',
			  'visible': True},
			 {'defvalue': '',
			  'description': 'Lowlevel deny filter',
			  'key': 'DENY_FILTERS',
			  'maxvalue': '',
			  'minvalue': '',
			  'type': 1,
			  'value': '',
			  'visible': True}]

		@rtype: C{list} of C{dict}
		@returns: a list of configvalues
		"""

		flags = self.flags()
		categories = self.categories()

		lst = plutil._do(self, self.session.config_list(), plutil._ListCB(self, self.session.config_list_handle, None, False))
		self._items = {}
		for item in lst:
			textflags = []
			textcategory = None
			for bit, name in flags:
				if (1 << bit) & item['flags']:
					textflags.append(name)
			for bit, name in categories:
				if (1 << bit) == item['category']:
					textcategory = name
			item['flags_text'] = textflags
			item['category_text'] = textcategory
			self._items[item['key']] = item

		return lst
