"""
This module contains BGPObject implementation.

There is no need to import this module directly. Use the methods of
the L{Ruleset object <ruleset.Ruleset>} to create and manipulate
BGPObjects.

    >>> o1 = r.object_add("/BGPObjects/Test1")
    >>> o1.add("2000@0") # ASPath@Origin
    >>>
    >>> o2 = r.object_add("/BGPObjects/Test2")
    >>> o.add("2000:2000") # Community:ASPath

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

import plobject
from thisversion import PLDB
import plutil

from packetlogic2.exceptions import PLDBUnsupportedInFirmware

def _verify_pos(pos):
	if pos < -1 or pos > 255:
		return False
	return True

class BGPObject(plobject.PLObject):
	"""
	BGPObject Implementation

	Object path: /BGPObjects

	@cvar Item: BGPObject Item implementation.
	"""

	__slots__ = ('_prop', '_pldb', '_callback', '_item_added')

	base_path = plutil.roConst("/BGPObjects", "Only the type portion of the path")
	type = plutil.roConst(PLDB.OBJECT_TYPE_BGPOBJECT, "Number representing the type of object")

	def _create_item(self, value):
		if not isinstance(value, basestring):
			raise TypeError("Item must be a string formatted as 'asnumber@pos' or 'asnumber:community'")

		if value.find("@") != -1:
			try:
				asnumber, pos = value.split("@", 1)
				item = self.Item(self, type=PLDB.BGPOBJECT_ITEM_TYPE_ASPATHPOS, asnumber=asnumber, value1=pos)
			except:
				raise ValueError("Not in the correct format, must be asnumber@pos")

			if not _verify_pos(item._prop.value1):
				raise ValueError("Position not a valid value")

		elif value.find(":") != -1:
			try:
				asnumber, community = value.split(":", 1)
				item = self.Item(self, type=PLDB.BGPOBJECT_ITEM_TYPE_COMMUNITY, asnumber=asnumber, value1=community)
			except:
				raise ValueError("Not in the correct format, must be asnumber:community")

		else:
			raise ValueError("Item must be a string formatted as 'asnumber@pos' or 'asnumber:community'")

		return item

	def add(self, value):
		"""
		Adds a BGPObject item to this BGPObject.

		@type value: C{str}

		@param value: This parameter describes the value of
		              the BGPObject item.  The string is
		              formatted as 'asnumber@position' or
		              'asnumber:community'.  Position must be
		              between -1 and 255. -1 = Last pos, 0 =
		              Any pos.

		@attention: The 'asnumber:community' format requires
		            firmware version 13.1.9 or newer.
		"""
		plobject.PLObject.add(self, value)


	class Item(plobject.PLObject.Item):
		"""
		BGPObject Item implementation.
		Should not be instantiated, use L{BGPObject.add} instead.

		@type type: C{int}
		@type asnumber: C{int}
		@type value1: C{int}
		"""
		__slots__ = ('_prop', '_parent')
		_Prop = plutil.Prop(type=int, asnumber=int, value1=int, *plobject._item_props)

		def __str__(self):
			if self._prop.type == PLDB.BGPOBJECT_ITEM_TYPE_ASPATHPOS:
				return "%u@%d" % (self._prop.asnumber, self._prop.value1)
			elif self._prop.type == PLDB.BGPOBJECT_ITEM_TYPE_ASREGEXP:
				raise ValueError("type == PLDB.BGPOBJECT_ITEM_TYPE_ASREGEXP is not yet implemented")
			elif self._prop.type == PLDB.BGPOBJECT_ITEM_TYPE_COMMUNITY:
				return "%u:%u" % (self._prop.asnumber, self._prop.value1)
			else:
				raise ValueError("Wrong item-type for this object")

		type = property(lambda self: self._prop.type, doc="Type (0=ASNumber@Pos, 1=ASregexp (not implemented), 2=ASNumber:Community)")
		asnumber = property(lambda self: self._prop.asnumber, doc="ASNumber")
		value1 = property(lambda self: self._prop.value1, doc="Position (-1 to 255) or Community number")
		pos = property(lambda self: self._prop.value1, doc="Position (-1 to 255)")
