"""
This module contains EnrichObject implementation.

There is no need to import this module directly. Use the methods of
the L{Ruleset object <ruleset.Ruleset>} to create and manipulate
EnrichObjects.

    >>> o = r.object_add("/EnrichObjects/Test")

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

import plobject
from thisversion import PLDB
import plutil

class EnrichObject(plobject.PLObject):
    """
    EnrichObject Implementation

    Object path: /EnrichObjects

    @cvar Item: EnrichObject Item implementation.
    """

    __slots__ = ('_prop', '_pldb', '_callback', '_item_added')

    base_path = plutil.roConst("/EnrichObjects", "Only the type portion of the path")
    type = plutil.roConst(PLDB.OBJECT_TYPE_ENRICHOBJECT, "Number representing the type of object")

    def _create_item(self, value):
        #
        #       item :: key op value optdefault
        #
        #        key :: '(netobject=' INT ')'
        #             | STR
        #
        #         op :: '=>'
        #             | '=-'
        #             | '=?'
        #             | '=+'
        #
        #      value :: '(netobject=' INT ')'
        #             | STR
        #
        # optdefault :: ''
        #             | ' || ' STR
        #
        #        STR :: [a-zA-Z0-9 _-]*
        #        INT :: [0-9]+
        #
        item = EnrichObject.Item(self)

        def parse_key(v):
            if v.startswith("(netobject="):
                # ex: '(netobject=10)'
                r = v.find(")")
                if r == -1:
                    raise ValueError("Couldn't parse item, bad netobject key")
                subval = v[len("(netobject="):r]
                if ':' in subval:
                    no, key = subval.split(":", 1)
                    netobjid = int(no)
                else:
                    netobjid = int(subval)
                    key = "1"
                item._prop.keytype = PLDB.ENRICHOBJECT_KEYTYPE_NETOBJECT_SUBSCRIBER
                item._prop.netobject = netobjid
                item._prop.key = key
                return r + 1
            elif v.startswith("(filename="):
                # ex: '(filename)'
                r = v.find(")")
                if r == -1:
                    raise ValueError("Couldn't parse item, bad filename key")
                item._prop.keytype = PLDB.ENRICHOBJECT_KEYTYPE_FILENAME
                item._prop.key = v[len("(filename="):r]
                return r + 1
            else:
                # ex: 'X-Forwarded-For'
                r = v.find("=")
                if r == -1:
                    raise ValueError("Couldn't parse item, bad key")
                item._prop.keytype = PLDB.ENRICHOBJECT_KEYTYPE_FIXED_STRING
                item._prop.key = v[:r]
                return r

        def parse_op(v):
            op = {
                "=>": PLDB.ENRICHOBJECT_SPOOF_OVERWRITE,
                "=-": PLDB.ENRICHOBJECT_SPOOF_REMOVE
            }.get(v[:2])
            if op is None:
                raise ValueError("Couldn't parse item, unknown operator %r" % v[:2])
            item._prop.spoof = op
            return 2

        def parse_val(v):
            if v.startswith("(netobject="):
                # ex: '(netobject=10)'
                r = v.find(")")
                if r == -1:
                    raise ValueError("Couldn't parse item, bad netobject value")
                subval = v[len("(netobject="):r]
                if ':' in subval:
                    no, value = subval.split(":", 1)
                    netobjid = int(no)
                else:
                    netobjid = int(subval)
                    value = "1"
                if item._prop.keytype == PLDB.ENRICHOBJECT_KEYTYPE_NETOBJECT_SUBSCRIBER and item._prop.netobject != netobjid:
                    raise ValueError("Key and value can't use different netobjects")
                item._prop.netobject = netobjid
                item._prop.value = value
                item._prop.valuetype = PLDB.ENRICHOBJECT_VALUETYPE_NETOBJECT_SUBSCRIBER
            elif v.startswith("(sc="):
                # ex: '(sc=myschema/subscriber)'
                r = v.find(")")
                if r == -1:
                    raise ValueError("Couldn't parse item, bad sessioncontext string")
                if '/' not in v:
                    raise ValueError("Couldn't parse item, bad sessioncontext string")
                item._prop.valuetype = PLDB.ENRICHOBJECT_VALUETYPE_SESSIONCONTEXT
                item._prop.value = v[1:-1]  # remove paranteses
            else:
                # ex: 'Some fixed string'
                item._prop.valuetype = PLDB.ENRICHOBJECT_VALUETYPE_FIXED_STRING
                item._prop.value = v

        if " || " in value:
            value, dft = value.rsplit(" || ", 1)
            item._prop.miss = PLDB.ENRICHOBJECT_MISS_USE_DEFAULT
            item._prop.defaultvalue = dft
        else:
            item._prop.miss = PLDB.ENRICHOBJECT_MISS_IGNORE

        off = parse_key(value)
        off += parse_op(value[off:])
        parse_val(value[off:])

        return item

    def add(self, value):
        """Add a EnrichObject Item to this EnrichObject

            >>> o = rs.object_add("/EnrichObjects/Forwarded for")
            >>> o.add("X-Forwarded-For=>(netobject=15624) || Unknown")
            >>> o.add("X-Via-PL=+I was here")

        @type value: C{str}
        @param value: A string that describes the value of the EnrichObject item
                      this should be formatted as I{key OP value}. Key
                      is either the name of the header, or one of the
                      special values '(netobject=N)' or
                      '(filename=A)'. OP is one of '=>' (replace), '=-'
                      (remove), '=?' (ignore) or '=+' (append). Value
                      is either a fixed string or the special value
                      '(netobject=N)'. If value ends with ' ||
                      something' that is used as default value.
        """
        plobject.PLObject.add(self, value)

    class Item(plobject.PLObject.Item):
        """
        EnrichObject Item implementation.
        Should not be instantiated, use L{EnrichObject.add} instead.

        @type field: C{int}
        @type value: C{str}
        """
        __slots__ = ('_prop', '_parent')
        _Prop = plutil.Prop(netobject=int, keytype=int, key=str, valuetype=int, value=str, spoof=int, miss=int, defaultvalue=str, *plobject._item_props)

        def __str__(self):
            if self.keytype == PLDB.ENRICHOBJECT_KEYTYPE_NETOBJECT_SUBSCRIBER:
                k = "(netobject=%s:%s)" % (self.netobject, self.key)
            elif self.keytype == PLDB.ENRICHOBJECT_KEYTYPE_FIXED_STRING:
                k = self.key
            elif self.keytype == PLDB.ENRICHOBJECT_KEYTYPE_FILENAME:
                k = "(filename=%s)" % self.key
            else:
                k = "(?(%d)=%s)" (self.keytype, self.key)

            if self.valuetype == PLDB.ENRICHOBJECT_VALUETYPE_NETOBJECT_SUBSCRIBER:
                v = "(netobject=%s:%s)" % (self.netobject, self.value)
            elif self.valuetype == PLDB.ENRICHOBJECT_VALUETYPE_FIXED_STRING:
                v = self.value
            elif self.valuetype == PLDB.ENRICHOBJECT_VALUETYPE_SESSIONCONTEXT:
                v = self.value
            else:
                v = "(?(%d)=%s)" (self.valuetype, self.value)

            if self.spoof == PLDB.ENRICHOBJECT_SPOOF_OVERWRITE:
                op = "=>"
            elif self.spoof == PLDB.ENRICHOBJECT_SPOOF_REMOVE:
                op = "=-"
            else:
                op = "=<%d>=" % self.spoof

            if self.miss == PLDB.ENRICHOBJECT_MISS_USE_DEFAULT:
                dft = " || %s" % self.defaultvalue
            else:
                dft = ""

            return "%s%s%s%s" % (k, op, v, dft)

        netobject = property(lambda self: self._prop.netobject, doc="")
        keytype = property(lambda self: self._prop.keytype, doc="")
        key = property(lambda self: self._prop.key, doc="")
        valuetype = property(lambda self: self._prop.valuetype, doc="")
        value = property(lambda self: self._prop.value, doc="")
        defaultvalue = property(lambda self: self._prop.defaultvalue, doc="")
        spoof = property(lambda self: self._prop.spoof, doc="")
        miss = property(lambda self: self._prop.miss, doc="")
