"""
@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

from thisversion import PLDB
import plutil

from packetlogic2.exceptions import PLDBUnsupportedInFirmware

def _add(pldb, obj_type, name, parent, parent_name):
	cid = pldb.session.object_add(obj_type, name, parent)
	id, tstamp = plutil._do(pldb, cid, plutil._AddCB(pldb, pldb.session.object_add_handle))
	res = pldb.type_to_obj[obj_type](
		id = id,
		name = name,
		path = parent_name,
		fullpath = parent_name + "/" + name,
		parent = parent,
		creator = pldb.user,
		modified_by = pldb.user,
		creation_date = tstamp,
		modification_date = tstamp,
		)
	res._pldb = pldb
	return res

def _remove(pldb, obj):
	cid = pldb.session.object_remove(obj.type, obj.id)
	return plutil._do(pldb, cid, plutil._AddCB(pldb, pldb.session.object_remove_handle))

def _unlink(pldb, obj):
	cid = pldb.session.object_unlink(obj.type, obj.id)
	return plutil._do(pldb, cid, plutil._Callback(pldb))

def _list(pldb, obj_type):
	### @raises RuntimeError("Unknown object item type"): for invalid type of object. Should never happen unless e.g two api-versions are mixed up.
	cid = pldb.session.object_list(obj_type)
	return plutil._do(pldb, cid, plutil._ListCB(pldb, pldb.session.object_list_handle, pldb.type_to_obj[obj_type]))

_item_props = ('id', 'creator', 'creation_date', 'modified_by', 'modification_date')


class PLObjectItem (object):
	"""
	Generic Object Item.
	Will only be used as superclass to other Object types Items.

	@type id: C{int}
	@type creator: C{str}
	@type creation_date: C{str}
	@type modified_by: C{str}
	@type modification_date: C{str}

	"""
	__slots__ = ('_prop', '_parent')
	_Prop = plutil.Prop(*_item_props)

	def __init__(self, parent, **props):
		"""
		Do not instantiate this manually. Use the add method of the object.
		"""
		self._parent = parent
		self._prop = self._Prop(**props)

	def __repr__(self):
		return "'" + self.__str__() + "'"

	def __str__(self):
		return "<Generic Object item>"

	def __eq__(self, other):
		if other.__class__ == str:
			if other == self.__str__():
				return True
			else:
				# Some objects can be created using different strings that
				# is not nessecarilly the same as represents it, so try
				# to create an object of our type with this string and
				# compare its representation, if successfull.
				try:
					other_item = self._parent._create_exclude(other)
				except:
					return False

				return str(other_item) == self.__str__()

		elif other.__class__ == self.__class__:
			return other.id == self._prop.id

		elif other.__class__ == int:
			return other == self._prop.id

		else:
			return False

	def __ne__(self, other):
		return not self.__eq__(other)

	def _modified(self):
		""" The property 'modified' is undocumented and deprecated, use 'modification_date' instead."""
		import warnings
		warnings.warn("The property 'modified' is undocumented and deprecated, use 'modification_date' instead.",
			      DeprecationWarning, stacklevel=2)
		return self._prop.modification_date

	def _created(self):
		""" The property 'created' is undocumented and deprecated, use 'creation_date' instead."""
		import warnings
		warnings.warn("The property 'created' is undocumented and deprecated, use 'creation_date' instead.",
			      DeprecationWarning, stacklevel=2)
		return self._prop.creation_date

	modified = property(_modified)
	created = property(_created)

	id = property(lambda self: self._prop.id, doc="Unique number used internally to identify this item.")
	creator = property(lambda self: self._prop.creator, doc="Username of this items creator.")
	creation_date = property(lambda self: self._prop.creation_date, doc="Date and time when item was created.")
	modified_by = property(lambda self: self._prop.modified_by, doc="Username that last modified this item.")
	modification_date = property(lambda self: self._prop.modification_date, doc="Date and time when this item was last modified.")


_Prop_PLObject = plutil.Prop('name', 'path', 'fullpath', 'parent', 'id', 'creation_date',
			     'creator', 'modification_date', 'modified_by', 'referenced',
			     items=list, visible=bool)
class PLObject(object):
	"""
	Generic PacketLogic Object class. This should be inherited
	from Object implementations. You should probably read the
	documentation for one of the Subclasses instead of this one.

	@type base_path: C{str}
	@type type: C{int}
	
	@type creation_date: C{str}
	@type creator: C{str}
	@type fullpath: C{str}
	@type id: C{int}
	@type items: read-only C{list}
	@type modification_date: C{str}
	@type modified_by: C{str}
	@type name: C{str}
	@type parent: C{int}
	@type path: C{str}
	@type is_global: C{boolean}
	@type visible: C{boolean}
	@type machineid: str
	
	@cvar Item: Generic Object Item implementation.
	@cvar __slots__:
	"""
	__slots__ = ('_prop', '_pldb', '_callback', '_item_added')
	_may_exclude = False # True if this object supports the exclude flag
	base_path = plutil.roConst("/Objects", "Only the type portion of the path")
	type = plutil.roConst(0, "Number representing the type of object")

	Item = PLObjectItem

	def __init__(self, name="Unnamed", id=0, parent=0, path="/", fullpath=None, **more_props):
		self._pldb = None
		self._prop = _Prop_PLObject(name=name, parent=parent, id=id, path=path, fullpath=fullpath, **more_props)
		
	def _make_path(self, iddict):
		""" This function will create self.path and self.fullpath """
		if self._prop.fullpath:
			return

		if self._prop.parent == 0:
			self._prop.path = self.base_path
		else:
			parent = iddict[self._prop.parent]
			if not parent.fullpath:
				parent._make_path(iddict)
			self._prop.path = parent.fullpath

		self._prop.fullpath = self._prop.path + "/" + self._prop.name

	def __repr__(self):
		return "<PacketLogic %s '%s' '%s'>" % (self.__class__.__name__, self._prop.name, self._prop.path)

	def _create_item(self, value):
		""" 
		This function must be overridden in object implementation
		"""
		raise NotImplementedError

	def set_visible(self, visible):
		"""
		Set or reset the visible flag on the current object.
		
		@type visible: C{bool}
		@param visible: True or False value for the flag.

		@raises ValueError: If the rule is global.
		"""
		self._api(None, self._pldb.session.object_set_visible, visible)
		
		self._prop.visible = bool(visible)

		self._pldb._invalidate_cache(self.type)

	def rename(self, newname):
		""" 
		Rename the current object

		I{destroys netobject cache}

		@type newname: C{str}
		@param newname: The name that the object should have after rename process

		@raises ValueError: If the rule is global.
		"""

		if newname == "":
			raise ValueError("Can't rename object to ''")

		self._api(self._pldb.session.object_rename_handle, self._pldb.session.object_rename, newname)

		# Will also update self.name and self.fullpath
		self._pldb._objectcache_rename(self, newname)

	def _split_pathstr(self, pathstr):

		obj = None
		path = None

		if len(pathstr) > 1:
			s = pathstr.split("/")
			if len(s) > 1:
				obj = s[1]

			if len(s) > 2:
				path = s[2:]

		return (obj, path)

	def move(self, newparent):
		"""
		Moves a object to somewhere else in the tree

		I{destroys the cache}

		@type newparent: C{str}
		@param newparent: a path where the new object should reside. Must
			start with the same objecttype. I.e. you can't move a NetObject to
			VlanIdObject.

		@raises ValueError: If the rule is global.
		@raises ValueError: If you specify an invalid path.
		"""

		if isinstance(newparent, basestring):

			# Extract object type and path
			obj, path = self._split_pathstr(newparent)

			if obj != self._split_pathstr(self.fullpath)[0]:
				raise ValueError("Wrong object type")

			if path is None:
				target_id = 0
			else:
				po = self._pldb.object_get(newparent)

				if po is None:
					raise ValueError("%s isn't a valid path" % newparent)

				target_id = po.id

		elif isinstance(newparent, (int, long)):
			# Assume is new parent is 0 we're moving to root.
			if newparent == 0:
				target_id = 0
			else:
				po = self._pldb.object_find_id(self.base_path, newparent)
				target_id = po.id
		else:
			raise ValueError("argument must be either a str or a int")

		self._api(self._pldb.session.object_move_handle, self._pldb.session.object_move, target_id)

		if target_id == 0:
			new_parent = self.base_path
		else:
			new_parent = po.fullpath
			
		# Change cache for this and all child objects
		# This will also set the new fullpath and path property of the objects.
		self._pldb._objectcache_move(self, new_parent)

		self._prop.parent = target_id
		
	def _set_path(self, new_fullpath):
		""" Callback from Ruleset._objectcahce_move that updates path and
		fullpath properties when an object is moved.
		"""
		self._prop.fullpath = new_fullpath
		self._prop.path = new_fullpath[:new_fullpath.rfind("/")]

	def _create_exclude(self, value, exclude=False):
		""" Create an item and set the exclude flag if present."""
		
		if not self._may_exclude:
			item = self._create_item(value)
			item._prop.update(id=0)

		else:
			if isinstance(value, basestring) and value[0:1] == "!":
				if exclude:
					raise ValueError('Using both the exclude argument and ! prefix is not supported.')
			
				exclude = True
				value = value[1:]

			item = self._create_item(value)
			item._prop.update(id=0, exclude=exclude)

		return item
		
	def add(self, value, exclude=False):
		"""
		Add an item to this Object. 

		@type value: C{str}
		@param value: A string with the value of the Item. The item type will be
			determined automatically. See each Objects documentation for format of
			the strings.

		@type exclude: C{boolean}
		@param exclude: If this is set to True the item will be a exclude
			Item instead of a include Item.

		@raises RuntimeError: B{"Unknown objecttype"}: for invalid type of object. Should never happen unless e.g two api-versions are mixed up.
		"""

		item = self._create_exclude(value, exclude)
		
		self._item_added = item
		self._api(self._add_item_callback, self._pldb.session.object_item_add, item)
		self._item_added = None

		list.append(self._prop.items, item) # add it to the local object too.

	def _add_item_callback(self, msg):
		(id, t) = self._pldb.session.object_item_add_handle(msg)
		self._item_added._prop.update(
			id = id,
			creation_date = t,
			modification_date = t,
			creator = self._pldb.user,
			modified_by = self._pldb.user,
			)
		return t
		
	def remove(self, item):
		"""
		Removes an Item from this Object

		@type item: L{Item} or C{str}
		@param item: Either you send in the Item class there or you send a string
			with the value. The B{first} Item that matches will be removed.

		@deprecated: The C{item} argument can also be an
		integer specifying the id number of the item to be
		removed. This is usage deprecated, instead use the
		remove_id method.
		"""

		if isinstance(item, (int, long)):
			import warnings
			warnings.warn("Remove by id is deprecated, use remove_id()",
				      DeprecationWarning, stacklevel=2)
		elif not isinstance(item, (basestring, self.Item)):
			raise TypeError("item must be item or string")

		i = self._prop.items[self._prop.items.index(item)]

		self._api(self._pldb.session.object_item_remove_handle, self._pldb.session.object_item_remove, i.id)

		list.remove(self._prop.items, item)

	def remove_id(self, id):
		"""
		Removes an Item from this Object by its idnumber.

			>>> i=o.add(7)
			>>> o.remove_id(i.id)

		@type item: C{int}
		@param item: ID-number of the item to remove
		"""
		for i in self._prop.items:
			if i.id == id:
				break
		else:
			raise ValueError("Item not found")
		self.remove(i)
		
	def _attr_set(self, key, value):
		self._api(None, self._pldb.session.object_attr_set,
			  key, value)

	def _attr_remove(self, key):
		self._api(None, self._pldb.session.object_attr_remove, key)

	def _attr_list(self):
		ci = self._pldb.session.object_attr_list(self.type, self._prop.id)
		c = plutil._SimpleListCB(self._pldb, self._pldb.session.object_attr_list_handle)
		c.list = []
		res = plutil._do(self._pldb, ci, c)
		return plutil.AttributesDict(res, self._attr_set, self._attr_remove)

	def _api(self, callback, function, *args):
		ci = function(self.type, self._prop.id, *args)
		self._callback = callback
		self._pldb._addcidcb(ci, self._handle)
		while self._callback is not False:
			self._pldb._process()
		
	def _handle(self, msg):
		if msg.get_type() == PLDB.MSG_TYPE_OK:
			self._callback = False
			return True
		
		if msg.get_type() == PLDB.MSG_TYPE_DATA and self._callback:
			self._prop.modification_date = self._callback(msg)
		
	def _machineid(self):
		"""
		Deprecated method to access 'systemid'.

		@deprecated: use 'systemid' instead.
		"""
		import warnings
		warnings.warn("Use of property 'machineid' is deprecated, use 'systemid' instead.",
			      DeprecationWarning, stacklevel=2)
		return self.systemid
	
	def _modified(self):
		"""
		Deprecated method to access 'modification_date'.

		@deprecated: use 'modification_date' instead.
		"""
		import warnings
		warnings.warn("The property 'modified' is undocumented and deprecated, use 'modification_date' instead.",
			      DeprecationWarning, stacklevel=2)
		return self._prop.modification_date
	
	machineid = property(_machineid)
	modified = property(_modified)

	_name_doc = """
	Name of the object.

	This is the name of the object, excluding path. Note that
	name is utf-8 encoded data in a string, to get a unicode
	string use C{o.name.decode("utf-8")}.

	To change the name of an object use the L{rename} method.

	@type: str
	"""
	name = property(lambda self: self._prop.name, doc=_name_doc)
	path = property(lambda self: self._prop.path, doc="Complete path to where the object is excluding the object name.")
	fullpath = property(lambda self: self._prop.fullpath, doc="Complete path including name of object")
	parent = property(lambda self: self._prop.parent, doc="id of parent object (or zero)")
	_id_doc = """
	Database id of object.

	The unique (for this object type) id number used to refer to the object in rule conditions.
	"""
	id = property(lambda self: self._prop.id, doc=_id_doc)
	creation_date = property(lambda self: self._prop.creation_date, doc="Date and time when this object was created")
	creator = property(lambda self: self._prop.creator, doc="Username that created this object")
	systemid = plutil._sysidprop("System ID of the PacketLogic this object belongs to")
	modification_date = property(lambda self: self._prop.modification_date, doc="Date and time when this object was last modified")
	modified_by = property(lambda self: self._prop.modified_by, doc="Username that last modified this object")
	is_global = property(lambda self: False, doc="True if this is object is not from the local system")
	visible = property(lambda self: self._prop.visible, doc="True if the object is visible in the GUI")
	_attributes_doc = """
	Arbitrary key/value attributes stored for this object.

	Attributes stored on objects doesn't affect ruleset
	evaluation, and are designed to be used by integrations to
	assign integration specific data to objects. The client does
	special handling for some attributes, such as "hidden" and
	"automatic".

	    >>> o.attributes
	    {}
	    >>> o.attributes['test'] = "Arbitrary string value"
	    >>> o.attributes
	    {'test': "Arbitrary string value"}

	@type: dict

	@raises PLDBUnsupportedInFirmware: If attributes is accessed without firmware support.

 	@attention: Attributes requires v12.2 firmware or newer.
	"""
	attributes = property(_attr_list, doc=_attributes_doc)
	
	items = plutil.roCopyList('items', "List of items associated with this object")

        _referenced = property(lambda self: self._prop.referenced)
